"""
    NEO JSON-RPC client utilities
    =============================

    This module defines utilities and shortcuts used by the NEO JSON-RPC client.

"""

import binascii
import copy
import re

from .constants import ContractParameterTypes


def is_hash256(s):
    """ Returns True if the considered string is a valid SHA256 hash. """
    if not s or not isinstance(s, str):
        return False
    return re.match('^[0-9A-F]{64}$', s.strip(), re.IGNORECASE)


def is_hash160(s):
    """ Returns True if the considered string is a valid RIPEMD160 hash. """
    if not s or not isinstance(s, str):
        return False
    if not len(s) == 40:
        return False
    for c in s:
        if (c < '0' or c > '9') and (c < 'A' or c > 'F') and (c < 'a' or c > 'f'):
            return False
    return True


def encode_invocation_params(params):
    """ Returns a list of paramaters meant to be passed to JSON-RPC endpoints. """
    final_params = []
    for p in params:
        if isinstance(p, bool):
            final_params.append({'type': ContractParameterTypes.BOOLEAN.value, 'value': p})
        elif isinstance(p, int):
            final_params.append({'type': ContractParameterTypes.INTEGER.value, 'value': p})
        elif is_hash256(p):
            final_params.append({'type': ContractParameterTypes.HASH256.value, 'value': p})
        elif is_hash160(p):
            final_params.append({'type': ContractParameterTypes.HASH160.value, 'value': p})
        elif isinstance(p, bytearray):
            final_params.append({'type': ContractParameterTypes.BYTE_ARRAY.value, 'value': p})
        elif isinstance(p, str):
            final_params.append({'type': ContractParameterTypes.STRING.value, 'value': p})
        elif isinstance(p, list):
            innerp = encode_invocation_params(p)
            final_params.append({'type': ContractParameterTypes.ARRAY.value, 'value': innerp})
    return final_params


def decode_invocation_result(result):
    """ Tries to decode the values embedded in an invocation result dictionary. """
    if 'stack' not in result:
        return result
    result = copy.deepcopy(result)
    result['stack'] = _decode_invocation_result_stack(result['stack'])
    return result


def _decode_invocation_result_stack(stack):
    stack = copy.deepcopy(stack)
    for value_dict in stack:
        if value_dict['type'] == 'Array':
            value_dict['value'] = _decode_invocation_result_stack(value_dict['value'])
        elif value_dict['type'] == 'ByteArray':
            value_dict['value'] = bytearray(binascii.unhexlify(value_dict['value'].encode('utf-8')))
    return stack



"""
    NEO JSON-RPC client
    ===================

    The NEO JSON-RPC client implements the JSON-RPC methods provided by the API interface of NEO
    nodes (see: http://docs.neo.org/en-us/node/api.html). NEO nodes can expose a JSON-RPC interface
    to allow clients to obtain blockchain data in order to ease the development of blockchain
    applications. The NEO JSON-RPC client also implements several client-specific methods or
    shortcuts in order to provide more high-level interfaces for interacting with the blockchain.

"""


__version__ = '0.1.0'


from .client import Client  # noqa: F401



"""
    NEO JSON-RPC client exceptions
    ==============================

    This module defines top-level exceptions that can be used by the NEO JSON-RPC client
    implementation.

"""


class JSONRPCError(Exception):
    """ Base exception for all exceptions that can be raised by the NEO JSON-RPC client. """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        return self.msg or super(JSONRPCError, self).__str__()


class TransportError(JSONRPCError):
    """ Raised when an error occurs related to the connection with the JSON-RPC server. """

    def __init__(self, msg, response):
        super(TransportError, self).__init__(msg)
        self.response = response


class ProtocolError(JSONRPCError):
    """ Raised when an error occurs related to the JSON-RPC protocol / the NEO JSON-RPC methods. """

    def __init__(self, msg, response, data=None):
        super(ProtocolError, self).__init__(msg)
        self.response = response
        self.data = data



from neojsonrpc.utils import (decode_invocation_result, encode_invocation_params, is_hash160,
                              is_hash256)


def test_is_hash256_helper_works():
    assert is_hash256('936a185caaa266bb9cbe981e9e05cb78cd732b0b3280eb944412bb6f8f8f07af')
    assert not is_hash256('dummy')
    assert not is_hash256('98c615784ccb5fe5936fbc0cbe9dfdb408d92f0f')


def test_is_hash160_helper_works():
    assert is_hash160('98c615784ccb5fe5936fbc0cbe9dfdb408d92f0f')
    assert not is_hash160('dummy')
    assert not is_hash160('936a185caaa266bb9cbe981e9e05cb78cd732b0b3280eb944412bb6f8f8f07af')
    assert not is_hash160('98c615784ccb5fe5936fbc0Ébe9dfdb408d92f0f')


class TestEncodeInvocationParamsHelper:
    def test_can_encode_a_boolean(self):
        assert encode_invocation_params([True, False, ]) == \
            [{'type': 'Boolean', 'value': True}, {'type': 'Boolean', 'value': False}, ]

    def test_can_encode_an_integer(self):
        assert encode_invocation_params([42, ]) == [{'type': 'Integer', 'value': 42}, ]

    def test_can_encode_a_hash256(self):
        assert encode_invocation_params(
            ['936a185caaa266bb9cbe981e9e05cb78cd732b0b3280eb944412bb6f8f8f07af', ]) == \
            [{'type': 'Hash256',
              'value': '936a185caaa266bb9cbe981e9e05cb78cd732b0b3280eb944412bb6f8f8f07af'}, ]

    def test_can_encode_a_hash160(self):
        assert encode_invocation_params(
            ['98c615784ccb5fe5936fbc0cbe9dfdb408d92f0f', ]) == \
            [{'type': 'Hash160', 'value': '98c615784ccb5fe5936fbc0cbe9dfdb408d92f0f'}, ]

    def test_can_encode_a_byte_array(self):
        assert encode_invocation_params([bytearray('test'.encode('utf-8')), ]) == \
            [{'type': 'ByteArray', 'value': bytearray('test'.encode('utf-8'))}, ]

    def test_can_encode_a_string(self):
        assert encode_invocation_params(['Hello', ]) == [{'type': 'String', 'value': 'Hello'}, ]

    def test_can_encode_an_array(self):
        assert encode_invocation_params([[1, 3, 6], ]) == \
            [{'type': 'Array',
              'value': [{'type': 'Integer', 'value': 1}, {'type': 'Integer', 'value': 3},
                        {'type': 'Integer', 'value': 6}]}]


class TestDecodeInvocationResultHelper:
    def test_do_nothing_if_stack_results_are_not_present(self):
        result = {
            'gas_consumed': '0.334',
            'script': '00000',
            'state': 'HALT, BREAK',
            'tx': '00000',
        }
        assert decode_invocation_result(result) == result

    def test_can_decode_byte_array_values(self):
        result = {
            'gas_consumed': '0.334',
            'script': '00000',
            'stack': [{'type': 'ByteArray', 'value': '68747470733a2f2f6e656f2e6f7267'}],
            'state': 'HALT, BREAK',
            'tx': '00000',
        }
        assert decode_invocation_result(result) == \
            {
                'gas_consumed': '0.334',
                'script': '00000',
                'stack': [{'type': 'ByteArray', 'value': bytearray(b'https://neo.org')}],
                'state': 'HALT, BREAK',
                'tx': '00000', }

    def test_can_decode_array_values(self):
        result = {
            'gas_consumed': '0.334',
            'script': '00000',
            'stack': [{'type': 'Array',
                       'value': [{'type': 'ByteArray',
                                  'value': '68747470733a2f2f6e656f2e6f7267'}]}],
            'state': 'HALT, BREAK',
            'tx': '00000',
        }
        assert decode_invocation_result(result) == \
            {
                'gas_consumed': '0.334',
                'script': '00000',
                'stack': [{'type': 'Array',
                           'value': [{'type': 'ByteArray',
                                      'value': bytearray(b'https://neo.org')}]}],
                'state': 'HALT, BREAK',
                'tx': '00000', }



from neojsonrpc import Client


class TestClient:
    def test_can_return_the_state_of_an_account(self):
        client = Client.for_testnet()
        account = client.get_account_state('ALn85kUVLWWZdCuZfvjwbmusWH9Hx43XuF')
        assert 'balances' in account
        assert account['balances']

    def test_can_return_the_state_of_an_asset(self):
        client = Client.for_testnet()
        asset = client.get_asset_state(
            'c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b')  # NEO
        assert asset['amount'] == '100000000'
        assert asset['available'] == '100000000'
        assert asset['precision'] == 0

    def test_can_return_the_best_block_hash(self):
        client = Client.for_testnet()
        best_block_hash = client.get_best_block_hash()
        assert len(best_block_hash[2:]) == 64

    def test_can_return_block_information_associated_with_a_block_hash(self):
        client = Client.for_testnet()
        block = client.get_block('9e38052ac0827e490b917bd69174cfa0fae9a7f77db18ed3cf2ee95922e0103d')
        assert block['hash'] == '0x9e38052ac0827e490b917bd69174cfa0fae9a7f77db18ed3cf2ee95922e0103d'
        assert block['index']

    def test_can_return_the_block_count(self):
        client = Client.for_testnet()
        assert client.get_block_count() > 0

    def test_can_return_the_block_hash_associated_with_an_index(self):
        client = Client.for_testnet()
        block_hash = client.get_block_hash(970238)
        assert block_hash == '0x631be7a0fc5bc684bd756003576d5b9ecccb096a636d5e86e951513d3bba537c'

    def test_can_return_system_fees_associated_with_a_block_hash(self):
        client = Client.for_testnet()
        fees = client.get_block_sys_fee(966045)
        assert fees == '716234'

    def test_can_return_the_connection_count(self):
        client = Client.for_testnet()
        connection_count = client.get_connection_count()
        assert connection_count

    def test_can_return_the_state_of_a_contract(self):
        client = Client.for_testnet()
        contract = client.get_contract_state('f9572c5b119a6b5775a6af07f1cef5d310038f55')
        assert contract['hash'] == '0xf9572c5b119a6b5775a6af07f1cef5d310038f55'

    def test_can_return_a_list_of_unconfirmed_transactions_associated_with_the_node(self):
        client = Client.for_testnet()
        mempool = client.get_raw_mem_pool()
        assert isinstance(mempool, list)

    def test_can_return_detailed_information_associated_with_a_transaction_hash(self):
        client = Client.for_testnet()
        tx = client.get_raw_transaction(
            'b96bd2b0070a4b25f47b390225da100b96ddcbe28c9af1f34cbb34462bf3db22')
        assert tx['txid'] == '0xb96bd2b0070a4b25f47b390225da100b96ddcbe28c9af1f34cbb34462bf3db22'
        assert tx['size'] == 145
        assert tx['blocktime'] == 1515210609

    def test_can_return_a_value_stored_in_the_storage_of_a_specific_contract(self):
        client = Client.for_testnet()
        val = client.get_storage('34af1b6634fcd7cfcff0158965b18601d3837e32', 'totalSupply')
        assert isinstance(val, bytearray)
        val = int.from_bytes(val, 'little')
        assert val == 100000000000
        assert not client.get_storage('34af1b6634fcd7cfcff0158965b18601d3837e32', 'dummyKey')

    def test_can_return_the_transaction_ouyput_information_associated_with_a_hash_and_index(self):
        client = Client.for_testnet()
        txout = client.get_tx_out(
            '6ed78234b2d4ac886f4f6246a19d237be38d6541e59d87b547987778314bb7af', 0)
        assert txout is None

    def test_can_return_node_peers(self):
        client = Client(host='52.15.48.60', port=8880)
        peers = client.get_peers()
        assert peers

    def test_can_invoke_a_contract(self):
        client = Client.for_testnet()
        result = client.invoke('34af1b6634fcd7cfcff0158965b18601d3837e32', ['symbol', []])
        assert result['state'] == 'HALT, BREAK'
        assert result['stack'] == [{'type': 'ByteArray', 'value': bytearray(b'FOX')}]

    def test_can_invoke_a_contract_with_a_specific_function(self):
        client = Client.for_testnet()
        result = client.invoke_function('34af1b6634fcd7cfcff0158965b18601d3837e32', 'symbol', [])
        assert result['state'] == 'HALT, BREAK'
        assert result['stack'] == [{'type': 'ByteArray', 'value': bytearray(b'FOX')}]

    def test_can_invoke_a_contract_using_a_script(self):
        client = Client.for_testnet()
        result = client.invoke_script('000673796d626f6c67327e83d30186b1658915f0cfcfd7fc34661baf34')
        assert result['state'] == 'HALT, BREAK'
        assert result['stack'] == [{'type': 'ByteArray', 'value': bytearray(b'FOX')}]

    def test_can_send_a_raw_transaction(self):
        client = Client.for_testnet()
        result = client.send_raw_transaction(
            '80000001195876cb34364dc38b730077156c6bc3a7fc570044a66fbfeeea56f71327e8ab0000029b7cffda'
            'a674beae0f930ebe6085af9093e5fe56b34a5c220ccdcf6efc336fc500c65eaf440000000f9a23e06f74cf'
            '86b8827a9108ec2e0f89ad956c9b7cffdaa674beae0f930ebe6085af9093e5fe56b34a5c220ccdcf6efc33'
            '6fc50092e14b5e00000030aab52ad93f6ce17ca07fa88fc191828c58cb71014140915467ecd359684b2dc3'
            '58024ca750609591aa731a0b309c7fb3cab5cd0836ad3992aa0a24da431f43b68883ea5651d548feb6bd3c'
            '8e16376e6e426f91f84c58232103322f35c7819267e721335948d385fae5be66e7ba8c748ac15467dcca06'
            '93692dac')
        assert result is False

    def test_can_validate_an_address(self):
        client = Client.for_testnet()
        assert client.validate_address('ASkaarMYjqdbcBpvLTFNXJhyonXR8K9Xx3')['isvalid']
        assert not client.validate_address('A=BSkaarMYjqdbcBpvLTFNXJhyonXR8K9Xx3')['isvalid']

    def test_can_invoke_a_contract_function_as_a_method(self):
        client = Client.for_testnet()
        result = client.contract('34af1b6634fcd7cfcff0158965b18601d3837e32').symbol()
        assert result['state'] == 'HALT, BREAK'
        assert result['stack'] == [{'type': 'ByteArray', 'value': bytearray(b'FOX')}]



"""
    NEO JSON-RPC client constants
    =============================

    This module defines top-level constants that can be used all tools or classes provided by
    neojsonrpc, such as the JSON-RPC client.

"""

from enum import Enum


class JSONRPCMethods(Enum):
    """ Defines the values of the NEO JSON-RPC methods. """

    # The 'getaccountstate' method allows to check account asset information according to an account
    # address.
    GET_ACCOUNT_STATE = 'getaccountstate'

    # The 'getassetstate' method allows to query the asset information, based on a specified asset
    # number.
    GET_ASSET_STATE = 'getassetstate'

    # The 'getbestblockhash' method allows to get the hash of the tallest block in the main chain.
    GET_BEST_BLOCK_HASH = 'getbestblockhash'

    # The 'getblock' method allows to fetch the block information associated with a specific hash
    # value or block index (block height).
    GET_BLOCK = 'getblock'

    # The 'getblockcount' method allows to get the number of blocks in the chain.
    GET_BLOCK_COUNT = 'getblockcount'

    # The 'getblockhash' method allows to get the hash associated with a specific block index.
    GET_BLOCK_HASH = 'getblockhash'

    # The 'getblocksysfee' method allows to get the system fees of a block identified by a specific
    # block index.
    GET_BLOCK_SYS_FEE = 'getblocksysfee'

    # The 'getconnectioncount' method allows to get the current number of connections for the
    # considered node.
    GET_CONNECTION_COUNT = 'getconnectioncount'

    # The 'getcontractstate' method allows to query contract information associated with a specific
    # script hash.
    GET_CONTRACT_STATE = 'getcontractstate'

    # The 'getrawmempool' method allows to get a list of unconfirmed transactions in memory.
    GET_RAW_MEM_POOL = 'getrawmempool'

    # The 'getrawtransaction' method allows to get detailed information associated with a specific
    # transaction hash.
    GET_RAW_TRANSACTION = 'getrawtransaction'

    # The 'getstorage' method allows to get a value stored in the storage associated with a specific
    # contract script hash and a specific key.
    GET_STORAGE = 'getstorage'

    # The 'gettxout' method allows to get the transaction output information corresponding to a
    # specified hash and index.
    GET_TX_OUT = 'gettxout'

    # The 'getpeers' method allows to get a list of nodes that the considered node is currently
    # connected/disconnected from.
    GET_PEERS = 'getpeers'

    # The 'getversion' method allows to get version information about the considered node.
    GET_VERSION = 'getversion'

    # The 'invoke' method allows to invoke a contract (associated with a specific script hash) with
    # potential parameters and to return the result. It should be noted that calling this method
    # does not affect the blockchain in any way: the targetted contract is executed with the given
    # parameters but the contract's storage remain unchanged.
    INVOKE = 'invoke'

    # The 'invokefunction' method allows to invoke a contract (associated with a specific script
    # hash) by specifying an operation and potential parameters. It should be noted that calling
    # this method does not affect the blockchain in any way: the targetted contract is executed with
    # the given parameters but the contract's storage remain unchanged.
    INVOKE_FUNCTION = 'invokefunction'

    # The 'invokescript' method allows to invoke a specific script that'll be run through the VM and
    # to get the corresponding result. This method is to test VM script as if they were ran on the
    # blockchain at a specific point in time. This RPC call does not affect the blockchain in any
    # way.
    INVOKE_SCRIPT = 'invokescript'

    # The 'sendrawtransaction' method allows to broadcast a transaction over the NEO network.
    SEND_RAW_TRANSACTION = 'sendrawtransaction'

    # The 'validateaddress' method allows to validate if a string is a valid NEO address.
    VALIDATE_ADDRESS = 'validateaddress'


class ContractParameterTypes(Enum):
    """ Defines the contract parameter types supported by the NEO JSON-RPC endpoints. """

    BOOLEAN = 'Boolean'
    INTEGER = 'Integer'
    HASH160 = 'Hash160'
    HASH256 = 'Hash256'
    BYTE_ARRAY = 'ByteArray'
    STRING = 'String'
    ARRAY = 'Array'



import unittest.mock

import pytest
from requests.exceptions import HTTPError

from neojsonrpc import Client
from neojsonrpc.exceptions import ProtocolError, TransportError


class TestClient:
    @unittest.mock.patch('requests.Session.post')
    def test_raises_a_transport_error_if_an_unsuccessful_is_sent_back_from_the_rpc_server(
            self, mocked_post):
        mocked_response = unittest.mock.Mock(status_code=500, content='ERROR')
        mocked_response.raise_for_status.side_effect = HTTPError(response=mocked_response)
        mocked_post.return_value = mocked_response
        client = Client.for_testnet()
        with pytest.raises(TransportError):
            client.get_block_count()

    @unittest.mock.patch('requests.Session.post')
    def test_raises_a_protocol_error_if_a_response_cannot_be_deserialized(self, mocked_post):
        mocked_response = unittest.mock.Mock(status_code=200, content='BAD')
        mocked_response.json.side_effect = ValueError()
        mocked_post.return_value = mocked_response
        client = Client.for_testnet()
        with pytest.raises(ProtocolError):
            client.get_block_count()

    @unittest.mock.patch('requests.Session.post')
    def test_raises_a_protocol_error_if_an_error_is_present_in_the_response(
            self, mocked_post):
        mocked_response = unittest.mock.Mock(status_code=200, content='{}')
        mocked_response.json.return_value = {'error': {'message': 'ERROR'}}
        mocked_post.return_value = mocked_response
        client = Client.for_testnet()
        with pytest.raises(ProtocolError):
            client.get_block_count()

    @unittest.mock.patch('requests.Session.post')
    def test_raises_a_protocol_error_if_result_data_is_not_present_in_the_response(
            self, mocked_post):
        mocked_response = unittest.mock.Mock(status_code=200, content='{}')
        mocked_response.json.return_value = {}
        mocked_post.return_value = mocked_response
        client = Client.for_testnet()
        with pytest.raises(ProtocolError):
            client.get_block_count()



"""
    NEO JSON-RPC client
    ===================

    This module defines the ``Client`` class allowing to interact with the JSON-RPC endpoints.

"""

import binascii
import json

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError

from .constants import JSONRPCMethods
from .exceptions import ProtocolError, TransportError
from .utils import decode_invocation_result, encode_invocation_params


class Client:
    """ The NEO JSON-RPC client class. """

    def __init__(self, host=None, port=None, tls=False, http_max_retries=None):
        # Initializes attributes related to the client settings (host, port, etc).
        self.host = host or 'localhost'
        self.port = port or 30333
        self.tls = tls
        self.session = requests.Session()
        self.session.mount(self.host, HTTPAdapter(max_retries=http_max_retries or 3))

        # Initializes an "ID counter" that'll be used to forge each request to the JSON-RPC
        # endpoint. The "id" parameter is "required" in order to help clients sort responses out.
        # In the case of the current client, we'll just ensure that this value gets incremented
        # after each request made to the JSON-RPC endpoint.
        self._id_counter = 0

    @classmethod
    def for_mainnet(cls):
        """ Creates a ``Client`` instance for use with the NEO Main Net. """
        return cls(host='seed3.neo.org', port=10331, tls=True)

    @classmethod
    def for_testnet(cls):
        """ Creates a ``Client`` instance for use with the NEO Test Net. """
        return cls(host='seed3.neo.org', port=20331, tls=True)

    def contract(self, script_hash):
        """ Returns a ``ContractWrapper`` instance allowing to easily invoke contract functions.

        This method allows to invoke smart contract functions as if they were Python class instance
        methods. For example:

        .. code-block:: python

            >>> contract = client.contract('34af1b6634fcd7cfcff0158965b18601d3837e32')
            >>> contract.symbol()
            {...}
            >>> contract.getBalance('<address>')
            {...}

        :param script_hash: contract script hash
        :type script_hash: str
        :return: :class:`ContractWrapper <ContractWrapper>` object
        :rtype: neojsonrpc.client.ContractWrapper

        """
        return ContractWrapper(self, script_hash)

    ####################
    # JSON-RPC METHODS #
    ####################

    def get_account_state(self, address, **kwargs):
        """ Returns the account state information associated with a specific address.

        :param address: a 34-bit length address (eg. AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz)
        :type address: str
        :return: dictionary containing the account state information
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_ACCOUNT_STATE.value, params=[address, ], **kwargs)

    def get_asset_state(self, asset_id, **kwargs):
        """ Returns the asset information associated with a specific asset ID.

        :param asset_id:
            an asset identifier (the transaction ID of the RegistTransaction when the asset is
            registered)
        :type asset_id: str
        :return: dictionary containing the asset state information
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_ASSET_STATE.value, params=[asset_id, ], **kwargs)

    def get_best_block_hash(self, **kwargs):
        """ Returns the hash of the tallest block in the main chain.

        :return: hash of the tallest block in the chain
        :rtype: str

        """
        return self._call(JSONRPCMethods.GET_BEST_BLOCK_HASH.value, **kwargs)

    def get_block(self, block_hash, verbose=True, **kwargs):
        """ Returns the block information associated with a specific hash value or block index.

        :param block_hash: a block hash value or a block index (block height)
        :param verbose:
            a boolean indicating whether the detailed block information should be returned in JSON
            format (otherwise the block information is returned as an hexadecimal string by the
            JSON-RPC endpoint)
        :type block_hash: str or int
        :type verbose: bool
        :return:
            dictionary containing the block information (or an hexadecimal string if verbose is set
            to False)
        :rtype: dict or str

        """
        return self._call(
            JSONRPCMethods.GET_BLOCK.value, params=[block_hash, int(verbose), ], **kwargs)

    def get_block_count(self, **kwargs):
        """ Returns the number of blocks in the chain.

        :return: number of blocks in the chain
        :rtype: int

        """
        return self._call(JSONRPCMethods.GET_BLOCK_COUNT.value, **kwargs)

    def get_block_hash(self, block_index, **kwargs):
        """ Returns the hash value associated with a specific block index.

        :param block_index: a block index (block height)
        :type block_index: int
        :return: hash of the block associated with the considered index
        :rtype: str

        """
        return self._call(JSONRPCMethods.GET_BLOCK_HASH.value, [block_index, ], **kwargs)

    def get_block_sys_fee(self, block_index, **kwargs):
        """ Returns the system fees associated with a specific block index.

        :param block_index: a block index (block height)
        :type block_index: int
        :return: system fees of the block, expressed in NeoGas units
        :rtype: str

        """
        return self._call(JSONRPCMethods.GET_BLOCK_SYS_FEE.value, [block_index, ], **kwargs)

    def get_connection_count(self, **kwargs):
        """ Returns the current number of connections for the considered node.

        :return: number of connections for the node
        :rtype: int

        """
        return self._call(JSONRPCMethods.GET_CONNECTION_COUNT.value, **kwargs)

    def get_contract_state(self, script_hash, **kwargs):
        """ Returns the contract information associated with a specific script hash.

        :param script_hash: contract script hash
        :type script_hash: str
        :return: dictionary containing the contract information
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_CONTRACT_STATE.value, [script_hash, ], **kwargs)

    def get_raw_mem_pool(self, **kwargs):
        """ Returns a list of unconfirmed transactions in memory associated with the node.

        :return: list of unconfirmed transaction hashes
        :rtype: list

        """
        return self._call(JSONRPCMethods.GET_RAW_MEM_POOL.value, **kwargs)

    def get_raw_transaction(self, tx_hash, verbose=True, **kwargs):
        """ Returns detailed information associated with a specific transaction hash.

        :param tx_hash: transaction hash
        :param verbose:
            a boolean indicating whether the detailed transaction information should be returned in
            JSON format (otherwise the transaction information is returned as an hexadecimal string
            by the JSON-RPC endpoint)
        :type tx_hash: str
        :type verbose: bool
        :return:
            dictionary containing the transaction information (or an hexadecimal string if verbose
            is set to False)
        :rtype: dict or str

        """
        return self._call(
            JSONRPCMethods.GET_RAW_TRANSACTION.value, params=[tx_hash, int(verbose), ], **kwargs)

    def get_storage(self, script_hash, key, **kwargs):
        """ Returns the value stored in the storage of a contract script hash for a given key.

        :param script_hash: contract script hash
        :param key: key to look up in the storage
        :type script_hash: str
        :type key: str
        :return: value associated with the storage key
        :rtype: bytearray

        """
        hexkey = binascii.hexlify(key.encode('utf-8')).decode('utf-8')
        hexresult = self._call(
            JSONRPCMethods.GET_STORAGE.value, params=[script_hash, hexkey, ], **kwargs)
        try:
            assert hexresult
            result = bytearray(binascii.unhexlify(hexresult.encode('utf-8')))
        except AssertionError:
            result = hexresult
        return result

    def get_tx_out(self, tx_hash, index, **kwargs):
        """ Returns the transaction output information corresponding to a hash and index.

        :param tx_hash: transaction hash
        :param index:
            index of the transaction output to be obtained in the transaction (starts from 0)
        :type tx_hash: str
        :type index: int
        :return: dictionary containing the transaction output
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_TX_OUT.value, params=[tx_hash, index, ], **kwargs)

    def get_peers(self, **kwargs):
        """ Returns a list of nodes that the node is currently connected/disconnected from.

        :return: dictionary containing the nodes the current node is connected/disconnected from
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_PEERS.value, **kwargs)

    def get_version(self, **kwargs):
        """ Returns version information about the current node.

        :return: dictionary containing version information about the current node
        :rtype: dict

        """
        return self._call(JSONRPCMethods.GET_VERSION.value, **kwargs)

    def invoke(self, script_hash, params, **kwargs):
        """ Invokes a contract with given parameters and returns the result.

        It should be noted that the name of the function invoked in the contract should be part of
        paramaters.

        :param script_hash: contract script hash
        :param params: list of paramaters to be passed in to the smart contract
        :type script_hash: str
        :type params: list
        :return: result of the invocation
        :rtype: dictionary

        """
        contract_params = encode_invocation_params(params)
        raw_result = self._call(
            JSONRPCMethods.INVOKE.value, [script_hash, contract_params, ], **kwargs)
        return decode_invocation_result(raw_result)

    def invoke_function(self, script_hash, operation, params, **kwargs):
        """ Invokes a contract's function with given parameters and returns the result.

        :param script_hash: contract script hash
        :param operation: name of the operation to invoke
        :param params: list of paramaters to be passed in to the smart contract
        :type script_hash: str
        :type operation: str
        :type params: list
        :return: result of the invocation
        :rtype: dictionary

        """
        contract_params = encode_invocation_params(params)
        raw_result = self._call(
            JSONRPCMethods.INVOKE_FUNCTION.value, [script_hash, operation, contract_params, ],
            **kwargs)
        return decode_invocation_result(raw_result)

    def invoke_script(self, script, **kwargs):
        """ Invokes a script on the VM and returns the result.

        :param script: script runnable by the VM
        :type script: str
        :return: result of the invocation
        :rtype: dictionary

        """
        raw_result = self._call(JSONRPCMethods.INVOKE_SCRIPT.value, [script, ], **kwargs)
        return decode_invocation_result(raw_result)

    def send_raw_transaction(self, hextx, **kwargs):
        """ Broadcasts a transaction over the NEO network and returns the result.

        :param hextx: hexadecimal string that has been serialized
        :type hextx: str
        :return: result of the transaction
        :rtype: bool

        """
        return self._call(JSONRPCMethods.SEND_RAW_TRANSACTION.value, [hextx, ], **kwargs)

    def validate_address(self, addr, **kwargs):
        """ Validates if the considered string is a valid NEO address.

        :param hex: string containing a potential NEO address
        :type hex: str
        :return: dictionary containing the result of the verification
        :rtype: dictionary

        """
        return self._call(JSONRPCMethods.VALIDATE_ADDRESS.value, [addr, ], **kwargs)

    ##################################
    # PRIVATE METHODS AND PROPERTIES #
    ##################################

    def _call(self, method, params=None, request_id=None):
        """ Calls the JSON-RPC endpoint. """
        params = params or []

        # Determines which 'id' value to use and increment the counter associated with the current
        # client instance if applicable.
        rid = request_id or self._id_counter
        if request_id is None:
            self._id_counter += 1

        # Prepares the payload and the headers that will be used to forge the request.
        payload = {'jsonrpc': '2.0', 'method': method, 'params': params, 'id': rid}
        headers = {'Content-Type': 'application/json'}
        scheme = 'https' if self.tls else 'http'
        url = '{}://{}:{}'.format(scheme, self.host, self.port)

        # Calls the JSON-RPC endpoint!
        try:
            response = self.session.post(url, headers=headers, data=json.dumps(payload))
            response.raise_for_status()
        except HTTPError:
            raise TransportError(
                'Got unsuccessful response from server (status code: {})'.format(
                    response.status_code),
                response=response)

        # Ensures the response body can be deserialized to JSON.
        try:
            response_data = response.json()
        except ValueError as e:
            raise ProtocolError(
                'Unable to deserialize response body: {}'.format(e), response=response)

        # Properly handles potential errors.
        if response_data.get('error'):
            code = response_data['error'].get('code', '')
            message = response_data['error'].get('message', '')
            raise ProtocolError(
                'Error[{}] {}'.format(code, message), response=response, data=response_data)
        elif 'result' not in response_data:
            raise ProtocolError(
                'Response is empty (result field is missing)', response=response,
                data=response_data)

        return response_data['result']


class ContractWrapper:
    """ Strategy class allowing to provide a high-level interface for invoking smart contracts. """

    def __init__(self, client, script_hash):
        self.client = client
        self.script_hash = script_hash

    def __getattribute__(self, attr):
        try:
            return super(ContractWrapper, self).__getattribute__(attr)
        except AttributeError:
            client = super(ContractWrapper, self).__getattribute__('client')
            script_hash = super(ContractWrapper, self).__getattribute__('script_hash')
            return ContractFunctionWrapper(client, script_hash, attr)


class ContractFunctionWrapper:
    """ Strategy class allowing to easily invoke smart contract functions. """

    def __init__(self, client, script_hash, funcname):
        self.client = client
        self.script_hash = script_hash
        self.funcname = funcname

    def __call__(self, *args):
        return self.client.invoke_function(self.script_hash, self.funcname, args)
